import torch

class DataCollatorWithPadding(object):
    r''' Will pad or trim examples to the appropriate length.
    '''
    def __init__(self, pad_token_id, ignore_index, max_seq_len):
        self.pad_token_id = pad_token_id
        self.ignore_index = ignore_index
        self.max_seq_len = max_seq_len

    def __call__(self, instances):
        input_ids, labels = tuple([torch.tensor(instance[key][0:self.max_seq_len]) for instance in instances] for key in ("input_ids", "labels"))
        batch = {}
        
        batch['input_ids'] = torch.nn.utils.rnn.pad_sequence(input_ids, batch_first=True, padding_value=self.pad_token_id) 
        batch['labels'] = torch.nn.utils.rnn.pad_sequence(labels, batch_first=True, padding_value=self.ignore_index)
        batch['attention_mask'] = batch['input_ids'].ne(self.pad_token_id)
    
        return batch


import pandas as pd
import math
import os
import json
import hashlib
from tqdm.auto import tqdm
from opengpt import parsers, teachers
import logging
import random


def split_csv_by_max_len(datasets, max_len, tokenizer, base_path):
    r''' Given a tokenizer it will split the dataset (based on the `text` column) into max_len sequencse 
    '''
    for dataset in tqdm(datasets, desc='Datasets', total=len(datasets)):
        csv_path = dataset['path']
        name = dataset['name']

        nrows = None
        if dataset.get('nrows', -1) > 0:
            nrows = dataset['nrows']

        df = pd.read_csv(csv_path, nrows=nrows)
        cols = df.columns
        assert 'text' in cols, f'The CSV for dataset {name} has no "text" column.'

        new_data = [list(cols) + ['len', 'part']]
        for _, row in tqdm(df.iterrows(), desc=dataset['name'], total=len(df)):
            text = row['text']
            tokens = tokenizer.encode(text)

            for i in range(math.ceil(len(tokens) / max_len)):
                new_text = tokenizer.decode(tokens[i*max_len:(i+1)*max_len])
                new_data_row = [row[c] if c != 'text' else new_text for c in cols]
                new_data_row.append(len(tokens[i*max_len:(i+1)*max_len]))
                new_data_row.append(f'part_{i}')
                new_data.append(new_data_row)

        # Save
        new_df = pd.DataFrame(new_data[1:], columns=new_data[0])
        new_df.to_csv(os.path.join(base_path, name, 'data_split_by_length.csv'), index=False)
        logging.warning(f'{dataset["name"]}: length before vs after: {len(df)} vs {len(new_df)}\n')


def create_dataset_no_input(config):
    r''' This does not require an input dataset to generate a new dataset, only a prompt is needed
    '''
    prompt_db = json.load(open(config.path.prompt_db, 'rb'))
    raw_data_columns = ['id', 'raw_output', 'prompt_hash']
    raw_data = pd.DataFrame(None, columns=raw_data_columns)
    raw_data_path = os.path.join(config.base_path, config.name, f"raw_generated_data_for_{config.name}.csv")
    if os.path.exists(raw_data_path):
        raw_data = pd.read_csv(raw_data_path)
        logging.warning(f"Loading an existing openai generated dataset found at: {raw_data_path}" + 
                        f"There are already {len(raw_data)} rows in the that dataset, the generation will continue from where last left off. " + 
                        f"The script will also do all examples that were not done in the previous run.")


    teacher = getattr(teachers, f'ask_{config.teacher.name}')
    for prompt_config in config.prompts: 
        prompts = [prompt for prompt in prompt_db if prompt['hash'] in prompt_config['hashes']] # There must be one

        parameters = prompt_config.get('extra_parameters', {})

        for language in prompt_config.get('languages', ['English']):
            parameters['language'] = language
            logging.warning(f"\nStarting prompts: {prompt_config['hashes']}\n #Runs: {prompt_config['runs']}\nLanguage: {language}")
            for prompt in prompts:
                # If some examples exist already


                start = len(raw_data[raw_data.prompt_hash == prompt['hash']])
                for _ in tqdm(range(start, prompt_config['runs']), total=(prompt_config['runs'] - start)):
                    prompt_text_template = prompt['text']
                    prompt_text = prompt_text_template.format(**parameters)
                    try:
                        out = teacher(prompt_text, config)
                        new_data = pd.DataFrame([[len(raw_data), out, prompt['hash']]], columns=raw_data_columns)
                        raw_data = pd.concat([raw_data, new_data], ignore_index=True)

                        if len(raw_data) % config.data_generation_checkpoint_every == 0:
                            logging.warning("Checkpointing the generated dataset.")
                            raw_data.to_csv(raw_data_path, index=False)

                    except Exception as e:
                        logging.exception(e)
                        logging.warning(f"Skipping example for prompt: {prompt['hash']}\n")

    if raw_data is not None and len(raw_data) > 0:
        raw_data.to_csv(raw_data_path, index=False)

    return raw_data


def create_dataset(config):
    prompt_db = json.load(open(config.path.prompt_db, 'rb'))
    raw_data_columns = ['id', 'raw_output', 'dataset', 'language', 'run', 'prompt_hash', 'prompt_text_hash', 'context']
    raw_data = pd.DataFrame(None, columns=raw_data_columns)
    prepared_data = None
    raw_data_path = os.path.join(config.base_path, config.name, f"raw_generated_data_for_{config.name}.csv")
    prepared_data_path = os.path.join(config.base_path, config.name, f"prepared_generated_data_for_{config.name}.csv")
    if os.path.exists(raw_data_path) and os.path.exists(prepared_data_path):
        raw_data = pd.read_csv(raw_data_path)
        prepared_data = pd.read_csv(prepared_data_path)
        logging.warning(f"Loading an existing openai generated dataset found at: \n{raw_data_path}\n and\n{prepared_data_path}\n" + 
                        f"There are already {len(raw_data)} rows in the that dataset, the generation will continue from where last left off. " + 
                        f"The script will also do all examples that were not done in the previous run.\n" + 
                        "***Take care that if prompt_config['random_prompt'] is set to true, it can produce unwanted results.\n\n")

    cnt = 0
    for prompt_config in config.prompts:
        prompts = [prompt for prompt in prompt_db if prompt['hash'] in prompt_config['hashes']] # There must be one
        teacher = getattr(teachers, f'ask_{config.teacher.name}')

        for run in range(prompt_config.get('runs', 1)):
            parameters = prompt_config.get('extra_parameters', {})
            extra_data_columns = prompt_config.get('extra_data_columns', [])

            for language in prompt_config.get('languages', ['English']):
                parameters['language'] = language
                logging.warning(f"\nStarting prompts: {prompt_config['hashes']}\nRun: {run}\nLanguage: {language}")
                for dataset_name in prompt_config['datasets']:
                    df = pd.read_csv(os.path.join(config.base_path, dataset_name, 'data_split_by_length.csv'))
                    for row_ind, row in tqdm(df.iterrows(), desc=dataset_name, total=len(df)):
                        # Set the context from the current row
                        parameters['context'] = row['text']
                        for col in extra_data_columns:
                            parameters[col] = row[col]
                        if prompt_config.get('random_prompt', False):
                            # This means for each example in the dataset we randomly select a prompt to be used, if False
                            #every example will run through every prompt
                            selected_prompts = [random.choice(prompts)]
                        else:
                            selected_prompts = prompts # Use all prompts sequentially
                        for prompt in selected_prompts:
                            prompt_text_template = prompt['text']
                            # Every prompt has its own parser
                            parser = getattr(parsers, prompt['parser'])
                            if len(str(row['text']).split(" ")) > config.teacher.min_len:
                                prompt_text = prompt_text_template.format(**parameters)
                                # The hash is of everything that is used to generate the output
                                h = hashlib.sha256(prompt_text.encode("utf-8"))
                                h.update(str(run).encode("utf-8"))
                                h = h.hexdigest()

                                # Only get the output if this was not done already
                                if h not in raw_data.prompt_text_hash.values:
                                    # Get output from OpenAI and parse using parser, the parser will append the parsed data onto the prepared_data CSV.
                                    try:
                                        openai_output = teacher(prompt_text, config)
                                        prepared_data = parser(data=openai_output, prepared_data=prepared_data, prompt_config=prompt_config, config=config, row=row, 
                                                               raw_data_id=len(raw_data), prompt_text=prompt_text) # ID is length of raw_data

                                        # Concat the current output to the data dataframe, only if not None
                                        if prepared_data is not None and len(prepared_data) > 0:
                                            new_data = pd.DataFrame([[len(raw_data), openai_output, dataset_name, language, run, prompt['hash'], h, parameters['context']]], 
                                                                    columns=raw_data_columns)
                                            raw_data = pd.concat([raw_data, new_data], ignore_index=True)
                                        if len(raw_data) % config.data_generation_checkpoint_every == 0:
                                            logging.warning("Checkpointing the generated dataset.")
                                            raw_data.to_csv(raw_data_path, index=False)
                                            prepared_data.to_csv(prepared_data_path, index=False)
                                    except Exception as e:
                                        logging.exception(e)
                                        logging.warning(f"Skipping example at position: {row_ind} for dataset: {dataset_name}\n")
    # Final save
    if raw_data is not None and prepared_data is not None and len(raw_data) > 0 and len(prepared_data) > 0:
        raw_data.to_csv(raw_data_path, index=False)
        prepared_data.to_csv(prepared_data_path, index=False)
    return raw_data, prepared_data 


def create_labels(examples, config, tokenizer):
    r''' This is used with a prepared HF dataset that is already tokenized. It will add labels
    so that only the AI generated parts (answers) will be trained on.
    '''
    
    user_token_id = tokenizer.vocab[config.special_tokens.user]
    ai_token_id = tokenizer.vocab[config.special_tokens.ai]
    # Everything written by an AI will be used for training, and everything by a user will be ignored

    examples['labels'] = []
    for i in range(len(examples['input_ids'])):
        labels = []
        ignore = True
        for tkn_id in examples['input_ids'][i]:
            if tkn_id == user_token_id:
                ignore = True
            elif tkn_id == ai_token_id:
                ignore = False
            
            if ignore:
                labels.append(config.train.ignore_index)
            else:
                labels.append(tkn_id)
        examples['labels'].append(labels)
    return examples


def pack_examples(examples, block_size, packing_type='partial'):
    r''' Used with a prepared HF dataset, will pack/group examples. Use with care, can mess up many things
    if the input is not formated properly (requires the <|eod|> token).
    
    packing_type: partial/full/no 
    '''
    # Concatenate all texts.
    if packing_type == 'partial':
        result = {k:[] for k in examples.keys()}
        _key = list(examples.keys())[0] # Take whichever key
        new_example = {k:[] for k in examples.keys()}

        for ind in range(len(examples[_key])):
            # Trim long sequences to block_size, this is required for partial packing
            example = {k:v[ind][0:block_size] for k,v in examples.items()}
            if len(new_example[_key]) + len(example[_key]) > block_size:
                result = {k:result[k] + [v] for k,v in new_example.items()}
                new_example = example 
            else:
                new_example = {k:new_example[k] + v for k,v in example.items()}
        #  Add the last example if there is something to add  
        if len(new_example[_key]) > 0:   
            result = {k:result[k] + [v] for k,v in new_example.items()}
    elif packing_type == 'full':
        # Full packing
        concatenated_examples = {k: sum(examples[k], []) for k in examples.keys()}
        total_length = len(concatenated_examples[list(examples.keys())[0]])
        total_length = (total_length // block_size) * block_size
        # Split by chunks of max_len.
        result = {
            k: [t[i : i + block_size] for i in range(0, total_length, block_size)]
            for k, t in concatenated_examples.items()
        }
    else:
        # Do nothing
        result = examples
    return result



from box import Box
import jsonpickle
import os
import yaml

class BaseConfig(object):
    def __init__(self, to_box=False):
        pass

    def _to_box(self):
        # Convert all dicts to boxes
        for key, val in self.__dict__.items():
            if isinstance(val, dict):
                self.__setattr__(key, Box(val))

    def _from_box(self):
        # Convert all dicts to boxes
        for key, val in self.__dict__.items():
            if isinstance(val, Box):
                self.__setattr__(key, val.to_dict())

    def save(self, save_path=None):
        r''' Save the config into a .json file
        Args:
            save_path (`str`):
                Where to save the created json file, if nothing we use the default from paths.
        '''
        if save_path is None:
            save_path = self.path.self

        # We want to save the dict here, not the whole class
        self._from_box()
        json_string = jsonpickle.encode({k:v for k,v in self.__dict__.items() if k != 'path'})

        with open(save_path, 'w') as f:
            f.write(json_string)
        self._to_box()

    @classmethod
    def load(cls, save_path):
        config = cls(to_box=False)
        # Read the jsonpickle string
        with open(save_path) as f:
            config_dict = jsonpickle.decode(f.read())
        config.merge_config(config_dict)
        config._to_box()
        return config

    def merge_config(self, config_dict):
        r''' Merge a config_dict with the existing config object.
        Args:
            config_dict (`dict`):
                A dictionary which key/values should be added to this class.
        '''
        for key in config_dict.keys():
            if key in self.__dict__ and isinstance(self.__dict__[key], dict):
                self.__dict__[key].update(config_dict[key])
            else:
                self.__dict__[key] = config_dict[key]


class Config(BaseConfig):
    r''' There are probably nicer ways to do this, but I like this one.
    '''
    def __init__(self, yaml_path):
        self.yaml_path = yaml_path
        self.load_yaml(yaml_path)

    def reload_yaml(self):
        self.load_yaml(self.yaml_path)

    def load_yaml(self, yaml_path):
        _config = yaml.safe_load(open(yaml_path, 'r'))
        self.to_box = True
        self.base_path = './'
        self.datasets = {}
        self.name = 'opengpt'

        for k,v in _config.items():
            self.__setattr__(k, v)
        # For fun, we will also keept the _config
        self._config = _config

        self.path = {'self': os.path.join(self.base_path, f'config_for_{self.name}.json')}
        if _config.get('static_paths', None):
            self.path.update(_config['static_paths'])

        if self.to_box:
            self._to_box()

            def create_dirs(paths):
                for path in paths:
                    if isinstance(path, str):
                        os.makedirs(os.path.dirname(path), exist_ok=True)
                    elif isinstance(path, dict):
                        create_dirs(path.values())
            create_dirs(self.path.values())
        
        # Create dirs for datasets, this is where all the data from one dataset will go
        for ds in self.datasets:
            os.makedirs(os.path.join(self.base_path, ds['name']), exist_ok=True)


r'''
Parsers are used to parse the output from a Teacher (OpenAI, Google, ...) into the right format. The purpose of the paraser is to
 parse the new output and append it to the prepared_data. Every parser will receive:
    - data: the new data output from a Teacher model
    - prepared_data: the dataset we are creating, in other words old data that was output by a parser
    - prompt_config: the prompt_config for the current prompt as a dictionary (taken from the .yaml file)
    - config: general config, ie the whole .yaml file as a python-box (can be used as a dictionary)
    - row: the row from the original CSV that was used for context to generate the `data`, can be empty given the use-case
    - raw_data_id: the ID of the `data` in the raw_data CSV (used to store the raw output from OpenAI)
    - prompt_text: the prepared prompt that was used to generate `data`

If we are running the paraser for the first time the `prepared_data` will be empty (None) and it is up to us to define how that prepared_data (e.g. CSV) should look. Every parser can have different columns depending on the use-case.

If the parser will output the final prepeared data that will be used for model training, it should append special tokens: config.special_tokens.[user, ai, eos, eod],
have a look at the functions below (e.g. csv_qa_parser).
'''

import pandas as pd
from io import StringIO
import re
import logging

def csv_qa_parser(data, prepared_data, prompt_config, config, row, raw_data_id, prompt_text):
    r''' Expects data in the CSV format, with the separator `;`, the dataframe has to have two columns: `Question`, `Answer`
    '''
    qa_pairs = None
    df = pd.read_csv(StringIO(data), sep=';')

    # Strip everything
    df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)

    ref_col = prompt_config.get('reference_column_to_append', None)
    if ref_col and row is not None and ref_col in row and row[ref_col]:
        # Means we want to append a reference at the end of each Answer
        to_append = f"\nReferences:\n- {row[ref_col]}"
        df['Answer'] = df['Answer'] + to_append
    df['Question'] += f' {config.special_tokens.eos}' # Every Q/A pair is independent
    df['Answer'] += f' {config.special_tokens.eos} {config.special_tokens.eod}'
    qa_pairs = [f'{config.special_tokens.user} {q.strip()} {config.special_tokens.ai} {a.strip()}' for q,a in df[['Question', 'Answer']].values]

    new_data = pd.DataFrame([[text, raw_data_id] for text in qa_pairs], columns=['text', 'raw_data_id'])
    if prepared_data is None:
        prepared_data = new_data
    else:
        prepared_data = pd.concat([prepared_data, new_data], ignore_index=True)

    return prepared_data


instruction_text = re.compile(r'Instruction:?(.*?)Input:', re.DOTALL)
input_text = re.compile(r'Input:?(.*?)Output:?', re.DOTALL)
output_text = re.compile(r'Output:?(.*?)$', re.DOTALL)
def task_parser(data, prepared_data, prompt_config, row, config, raw_data_id, prompt_text):
    r''' This parser can be used with prompts similar to Alpaca, it expects `data` in the following format:
        Task:
        Instruction:
        Input:
        Output:
        
        Task:
        Instruction:
        Input:
        Output:
    .
    .
    .
    '''
    tasks = re.split(r'[1-9 \.]*Task[:\s]*', str(data))
    st = config.special_tokens
    new_data = []
    for task in tasks:
        task = task.strip()
        ins = re.search(instruction_text, task).group(1).strip()
        inp = re.search(input_text, task).group(1).strip()
        out = re.search(output_text, task).group(1).strip()

        if inp:
            if inp.startswith('"'):
                inp = inp[1:]
            if inp.endswith('"'):
                inp = inp[:-1]
            if inp == '<noinput>':
                inp = ''
            else:
                inp = '\n' + str(inp)

        if ins and out:
            if inp in ins:
                new_data.append((len(prepared_data), f'{st.user} {ins} {st.eos} {st.ai} {out} {st.eos} {st.eod}', raw_data_id))
            else:
                new_data.append((len(prepared_data), f'{st.user} {ins}{inp} {st.eos} {st.ai} {out} {st.eos} {st.eod}', raw_data_id))
    
    new_data = pd.DataFrame(new_data, columns=['text', 'raw_data_id'])
    if prepared_data is None:
        prepared_data = new_data
    else:
        prepared_data = pd.concat([prepared_data, new_data], ignore_index=True)

    return prepared_data   


def simple_task_parser(data, prepared_data, prompt_config, row, config, raw_data_id, prompt_text):
    r''' This parser can be used with prompts similar to Alpaca, but that only have Instructions, it expects data :
        Task Number:
        Instruction:
        
        Task Number:
        Instruction:
        
    This parser is used as an intermediate, so the output is a csv with columns `text`, `instruction`, `raw_data_id`
    .
    .
    .
    '''
    tasks = [x.replace("Instruction:", "").strip() for x in re.split(r'[1-9 \.]*Task Number[:\s]*[\d\n]*', str(data)) if x.strip()]
    new_data = []
    for task in tasks:
        task = task.strip()
   
    new_data = pd.DataFrame([[[row['text']], task, raw_data_id] for task in tasks], columns=['text', 'instruction', 'raw_data_id'])
    if prepared_data is None:
        prepared_data = new_data
    else:
        prepared_data = pd.concat([prepared_data, new_data], ignore_index=True)

    return prepared_data   


def medical_conversation_parser(data, prepared_data, prompt_config, config, row, raw_data_id, prompt_text):
    r''' It expects data to be in form of a conversation, like:
        Patient: <some text>
        AI-Assistant: <some text>
        Patient: <some text>
        .
        .
        .
    The actor names 'Patient' and 'AI-Assistant" have to match exactlty 
    '''
    conversation = None

    # Merge the extractions into one conversation
    data = re.split(r'\s*(Patient\s*:|AI-Assistant\s*:)\s*', data)[1:]
    if len(data) > 0:
        conversation = ""
        to_append = None

        ref_col = prompt_config.get('reference_column_to_append', None)
        if ref_col and ref_col in row and row[ref_col]:
            # Means we want to append a reference at the end of each Answer
            to_append = f"\nReferences:\n- {row[ref_col]}"

        actor = None
        for message in data:
            message = message.strip()
            if message in ['Patient:', 'AI-Assistant:', 'Patient', 'AI-Assistant', 'Patient :', 'AI-Assistant :']:
                actor = message
            elif actor is not None: #TODO: Make this nicer
                if actor in ['Patient:', 'Patient :', 'Patient']:
                    conversation += f'{config.special_tokens.user} {message} {config.special_tokens.eos} '
                elif actor in ['AI-Assistant:', 'AI-Assistant :', 'AI-Assistant']:
                    conversation += f'{config.special_tokens.ai} {message}'
                    if to_append is not None and to_append:
                        conversation += to_append
                    conversation += f" {config.special_tokens.eos} "
        if conversation:
            conversation = conversation.strip() + f" {config.special_tokens.eod}"

    new_data = pd.DataFrame([[conversation, raw_data_id]], columns=['text', 'raw_data_id'])
    if prepared_data is None:
        prepared_data = new_data
    else:
        prepared_data = pd.concat([prepared_data, new_data], ignore_index=True)

    return prepared_data


def csv_ner_parser(data, prepared_data, prompt_config, config, row, raw_data_id, prompt_text):
    r''' Expects data in CSV format, using the `;` separator
    '''
    df = pd.read_csv(StringIO(data), sep=';', engine='python')
    df['raw_data_id'] = raw_data_id

    if prepared_data is None:
        prepared_data = df
    else:
        prepared_data = pd.concat([prepared_data, df], ignore_index=True)

    return prepared_data


import json
import hashlib
import os
import logging

def add_to_prompt_database(text, description, parser, database_path, force_replace=False):
    r''' The database is a simple json file where all the prompts are saved.
    '''
    if os.path.exists(database_path):
        logging.info(f"Loading db from: {database_path}")
        db = json.load(open(database_path, 'r'))
        hashes = set([prompt['hash'] for prompt in db])
    else:
        db = []
        hashes = set()

    # Good enough for what we need 
    h = hashlib.sha256(text.encode("utf-8")).hexdigest()[:10]
    if force_replace and h in hashes:
        logging.warning("Found an existing prompt with the same hash, it will be replaced with the new one.")
        # Remove the prompt with the hash as the current one
        db = [prompt for prompt in db if prompt['hash'] != h]
        hashes = set([prompt['hash'] for prompt in db])
    if h not in hashes:
        db.append({
                  'hash': h,
                  'text': text,
                  'description': description,
                  'parser': parser
                  })
        
        json.dump(db, open(database_path, 'w'), indent=2)
        logging.warning(f"Added prompt: {h}")
    else:
        logging.warning("The prompt is already in the database. It will not be added, you can use force_replace if you really want to add it.")

    return db


import setuptools
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info

with open("./README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="opengpt",
    version="0.0.5",
    author="w-is-h",
    author_email="w.kraljevic@gmail.com",
    description="OpenGPT a framework for producing grounded domain specific LLMs, and NHS-LLM a conversational model for healthcare made using OpenGPT.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cogstack/opengpt",
    packages=['opengpt'],
    install_requires=[
        'datasets>=2,<3',
        'transformers>=4.2,<5',
        'tiktoken>=0.3.2',
        'pandas',
        'openai',
        'numpy',
        'tqdm',
        'python-box',
        'jsonpickle',
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import openai

def ask_openai(prompt, config):
    response = openai.ChatCompletion.create(
        model = config.teacher.model,
        messages = [
            {"role": "user", "content": prompt},
        ]
    )

    message = None
    if response['choices'][0]['finish_reason'] == 'stop':
        message = response['choices'][0]['message']['content']

    return message


import logging

def add_tokens_to_model_and_tokenizer(config, tokenizer, model):
    ntkns = tokenizer.add_tokens(list(config.special_tokens.values()))
    logging.warning(f"Added: {ntkns} tokens to the tokenizer")
    if ntkns > 0:
        input_embeddings = model.get_input_embeddings().weight.data
        output_embeddings = model.get_output_embeddings().weight.data
        input_embeddings_avg = input_embeddings[:-ntkns].mean(dim=0, keepdim=True)
        output_embeddings_avg = output_embeddings[:-ntkns].mean(dim=0, keepdim=True)
        model.resize_token_embeddings(len(tokenizer))
        input_embeddings[-ntkns:] = input_embeddings_avg
        output_embeddings[-ntkns:] = output_embeddings_avg 
    
    # Set the eos and pad tokens properly
    tokenizer.add_special_tokens({"eos_token": config.special_tokens.eos, "pad_token": config.special_tokens.pad})
    model.config.eos_token_id = tokenizer.eos_token_id

    assert model.get_input_embeddings().num_embeddings == len(tokenizer)